@@ -8,6 +8,8 @@ import io
 import os
 
 import fitz
+import pathlib
+import pickle
 
 scriptdir = os.path.abspath(os.path.dirname(__file__))
 filename = os.path.join(scriptdir, "resources", "001003ED.pdf")
@@ -86,7 +88,7 @@ def test_annot_clean_contents():
     annot = page.add_highlight_annot((10, 10, 20, 20))
 
     # the annotation appearance will not start with command b"q"
-    assert annot._getAP().startswith(b"q") == False
+
 
     # invoke appearance stream cleaning and reformatting
     annot.clean_contents()
@@ -152,6 +154,7 @@ def test_bug1971():
         page = next(doc.pages())
         page.get_drawings()
         doc.close()
+        assert doc.is_closed
 
 def test_default_font():
     f = fitz.Font()
@@ -538,6 +541,12 @@ def test_2692():
                 alpha=False,
                 clip=fitz.Rect(0,0,10,10),
                 )
+        pix = dl.get_pixmap(
+                matrix=fitz.Identity,
+                #colorspace=fitz.csRGB,
+                alpha=False,
+                clip=fitz.Rect(0,0,10,10),
+                )
     
 
 def test_2596():
@@ -561,9 +570,67 @@ def test_2730():
     page = doc[0]
     s1 = set(page.get_text())  # plain text extraction
     s2 = set(page.get_text(sort=True))  # uses "blocks" extraction
+    s3 = set(page.get_textbox(page.rect))
     assert s1 == s2
+    assert s1 == s3
 
 
+def test_2553():
+    """Ensure identical output across text extractions."""
+    doc = fitz.open(f"{scriptdir}/resources/test_2553.pdf")
+    page = doc[0]
+
+    # extract plain text, build set of all characters
+    list1 = page.get_text()
+    set1 = set(list1)
+
+    # extract text blocks, build set of all characters
+    list2 = page.get_text(sort=True)  # internally uses "blocks"
+    set2 = set(list2)
+    
+    # extract textbox content, build set of all characters
+    list3 = page.get_textbox(page.rect)
+    set3 = set(list3)
+    
+    def show(l):
+        ret = f'len={len(l)}\n'
+        for c in l:
+            cc = ord(c)
+            if (cc >= 32 and cc < 127) or c == '\n':
+                ret += c
+            else:
+                ret += f' [0x{hex(cc)}]'
+        return ret
+    print(f'list1:\n{show(list1)}')
+    print(f'list2:\n{show(list2)}')
+    print(f'list3:\n{show(list3)}')
+    
+    # all sets must be equal
+    assert set1 == set2
+    assert set1 == set3
+
+    # With mupdf later than 1.23.4, this special page contains no invalid
+    # Unicodes.
+    #
+    if fitz.mupdf_version_tuple > (1, 23, 4):
+        print(f'Checking no occurrence of 0xFFFD, {fitz.mupdf_version_tuple=}.')
+        assert chr(0xFFFD) not in set1
+    else:
+        print(f'Checking occurrence of 0xFFFD, {fitz.mupdf_version_tuple=}.')
+        assert chr(0xFFFD) in set1
+
+def test_2553_2():
+    doc = fitz.open(f"{scriptdir}/resources/test_2553-2.pdf")
+    page = doc[0]
+
+    # extract plain text, ensure that there are no 0xFFFD characters
+    text = page.get_text()
+    if fitz.mupdf_version_tuple >= (1, 23, 7):
+        assert chr(0xfffd) not in text
+    else:
+        # Bug not fixed in MuPDF.
+        assert chr(0xfffd) in text
+
 def test_2635():
     """Rendering a page before and after cleaning it should yield the same pixmap."""
     doc = fitz.open(f"{scriptdir}/resources/test_2635.pdf")
@@ -572,4 +639,280 @@ def test_2635():
 
     page.clean_contents()  # clean page
     pix2 = page.get_pixmap()  # pixmap after cleaning
-    assert pix1.samples == pix2.samples  # assert equality
\ No newline at end of file
+    assert pix1.samples == pix2.samples  # assert equality
+
+
+def test_resolve_names():
+    """Test PDF name resolution."""
+    # guard against wrong PyMuPDF architecture version
+    if not hasattr(fitz.Document, "resolve_names"):
+        print("PyMuPDF version does not support resolving PDF names")
+        return
+    pickle_in = open(f"{scriptdir}/resources/cython.pickle", "rb")
+    old_names = pickle.load(pickle_in)
+    doc = fitz.open(f"{scriptdir}/resources/cython.pdf")
+    new_names = doc.resolve_names()
+    assert new_names == old_names
+
+def test_2777():
+    document = fitz.Document()
+    page = document.new_page()
+    print(page.mediabox.width)
+
+def test_2710():
+    doc = fitz.open(f'{scriptdir}/resources/test_2710.pdf')
+    page = doc.load_page(0)
+    
+    print(f'test_2710(): {page.cropbox=}')
+    print(f'test_2710(): {page.mediabox=}')
+    print(f'test_2710(): {page.rect=}')
+    
+    def numbers_approx_eq(a, b):
+        return abs(a-b) < 0.001
+    def points_approx_eq(a, b):
+        return numbers_approx_eq(a.x, b.x) and numbers_approx_eq(a.y, b.y)
+    def rects_approx_eq(a, b):
+        return points_approx_eq(a.bottom_left, b.bottom_left) and points_approx_eq(a.top_right, b.top_right)
+    def assert_rects_approx_eq(a, b):
+        assert rects_approx_eq(a, b), f'Not nearly identical: {a=} {b=}'
+                
+    blocks = page.get_text('blocks')
+    print(f'test_2710(): {blocks=}')
+    assert len(blocks) == 2
+    block = blocks[1]
+    rect = fitz.Rect(block[:4])
+    text = block[4]
+    print(f'test_2710(): {rect=}')
+    print(f'test_2710(): {text=}')
+    assert text == 'Text at left page border\n'
+    
+    assert_rects_approx_eq(page.cropbox, fitz.Rect(30.0, 30.0, 565.3200073242188, 811.9199829101562))
+    assert_rects_approx_eq(page.mediabox, fitz.Rect(0.0, 0.0, 595.3200073242188, 841.9199829101562))
+    print(f'test_2710(): {fitz.mupdf_version_tuple=}')
+    if fitz.mupdf_version_tuple < (1, 23, 5):
+        print(f'test_2710(): Not Checking page.rect and rect.')
+    elif fitz.mupdf_version_tuple < (1, 24.0):
+        print(f'test_2710(): Checking page.rect and rect.')
+        assert_rects_approx_eq(page.rect, fitz.Rect(0.0, 0.0, 535.3200073242188, 781.9199829101562))
+        assert_rects_approx_eq(rect, fitz.Rect(0.7872352600097656, 64.7560043334961, 124.85531616210938, 78.1622543334961))
+    else:
+        # 2023-11-05: Currently broken in mupdf master.
+        print(f'test_2710(): Not Checking page.rect and rect.')
+
+
+def test_2736():
+    """Check handling of CropBox changes vis-a-vis a MediaBox with
+       negative coordinates."""
+    doc = fitz.open()
+    page = doc.new_page()
+
+    # fake a MediaBox for demo purposes
+    doc.xref_set_key(page.xref, "MediaBox", "[-30 -20 595 842]")
+
+    assert page.cropbox == fitz.Rect(-30, 0, 595, 862)
+    assert page.rect == fitz.Rect(0, 0, 625, 862)
+
+    # change the CropBox: shift by (10, 10) in both dimensions. Please note:
+    # To achieve this, 10 must be subtracted from 862! yo must never be negative!
+    page.set_cropbox(fitz.Rect(-20, 0, 595, 852))
+
+    # get CropBox from the page definition
+    assert doc.xref_get_key(page.xref, "CropBox")[1] == "[-20 -10 595 842]"
+    assert page.rect == fitz.Rect(0, 0, 615, 852)
+
+    error = False
+    text = ""
+    try:  # check error detection
+        page.set_cropbox((-35, -10, 595, 842))
+    except Exception as e:
+        text = str(e)
+        error = True
+    assert error == True
+    assert text == "CropBox not in MediaBox"
+
+
+def test_subset_fonts():
+    """Confirm subset_fonts is working."""
+    if not hasattr(fitz, "mupdf"):
+        print("Not testing 'test_subset_fonts' in classic.")
+        return
+    text = "Just some arbitrary text."
+    arch = fitz.Archive()
+    css = fitz.css_for_pymupdf_font("ubuntu", archive=arch)
+    css += "* {font-family: ubuntu;}"
+    doc = fitz.open()
+    page = doc.new_page()
+    page.insert_htmlbox(page.rect, text, css=css, archive=arch)
+    doc.subset_fonts(verbose=True)
+    found = False
+    for xref in range(1, doc.xref_length()):
+        if doc.xref_is_font(xref):
+            if "+Ubuntu#20Regular" in doc.xref_object(xref):
+                found = True
+                break
+    assert found is True
+
+
+def test_2957_1():
+    """Text following a redaction must not change coordinates."""
+    # test file with redactions
+    doc = fitz.open(os.path.join(scriptdir, "resources", "test_2957_1.pdf"))
+    page = doc[0]
+    # search for string that must not move by redactions
+    rects0 = page.search_for("6e9f73dfb4384a2b8af6ebba")
+    # sort rectangles vertically
+    rects0 = sorted(rects0, key=lambda r: r.y1)
+    assert len(rects0) == 2  # must be 2 redactions
+    page.apply_redactions()
+
+    # reload page to finalize updates
+    page = doc.reload_page(page)
+
+    # the two string must retain their positions (except rounding errors)
+    rects1 = page.search_for("6e9f73dfb4384a2b8af6ebba")
+    rects1 = sorted(rects1, key=lambda r: r.y1)
+
+    assert page.first_annot is None  # make sure annotations have disappeared
+    for i in range(2):
+        r0 = rects0[i].irect  # take rounded rects
+        r1 = rects1[i].irect
+        assert r0 == r1
+
+
+def test_2957_2():
+    """Redacted text must not change positions of remaining text."""
+    doc = fitz.open(os.path.join(scriptdir, "resources", "test_2957_2.pdf"))
+    page = doc[0]
+    words0 = page.get_text("words")  # all words before redacting
+    page.apply_redactions()  # remove/redact the word "longer"
+    words1 = page.get_text("words")  # extract words again
+    assert len(words1) == len(words0) - 1  # must be one word less
+    assert words0[3][4] == "longer"  # just confirm test file is correc one
+    del words0[3]  # remove the redacted word from first list
+    for i in range(len(words1)):  # compare words
+        w1 = words1[i]  # word after redaction
+        bbox1 = fitz.Rect(w1[:4]).irect  # its IRect coordinates
+        w0 = words0[i]  # word before redaction
+        bbox0 = fitz.Rect(w0[:4]).irect  # its IRect coordinates
+        assert bbox0 == bbox1  # must be same coordinates
+
+def test_3070():
+    with fitz.open(os.path.abspath(f'{__file__}/../../tests/resources/test_3070.pdf')) as pdf:
+      links = pdf[0].get_links()
+      links[0]['uri'] = "https://www.ddg.gg"
+      pdf[0].update_link(links[0])
+      pdf.save(os.path.abspath(f'{__file__}/../../tests/test_3070_out.pdf'))
+
+def test_bboxlog_2885():
+    doc = fitz.open(os.path.abspath(f'{__file__}/../../tests/resources/test_2885.pdf'))
+    page=doc[0]
+    bbl = page.get_bboxlog()
+    bbl = page.get_bboxlog(layers=True)
+
+def test_3081():
+    '''
+    Check Document.close() closes file handles, even if a Page instance exists.
+    '''
+    path1 = os.path.abspath(f'{__file__}/../../tests/resources/1.pdf')
+    path2 = os.path.abspath(f'{__file__}/../../tests/test_3081-2.pdf')
+    path3 = os.path.abspath(f'{__file__}/../../tests/test_3081-3.pdf')
+    
+    rebased = hasattr(fitz, 'mupdf')
+    
+    import shutil
+    import sys
+    import traceback
+    shutil.copy2(path1, path2)
+    
+    def next_fd():
+        fd = os.open(path2, os.O_RDONLY)
+        os.close(fd)
+        return fd
+    
+    fd1 = next_fd()
+    document = fitz.open(path2)
+    page = document[0]
+    fd2 = next_fd()
+    document.close()
+    if rebased:
+        assert document.this is None
+        assert page.this is None
+    try:
+        document.page_count()
+    except Exception as e:
+        print(f'Received expected exception: {e}')
+        #traceback.print_exc(file=sys.stdout)
+        assert str(e) == 'document closed'
+    else:
+        assert 0, 'Did not receive expected exception.'
+    fd3 = next_fd()
+    try:
+        page.bound()
+    except Exception as e:
+        print(f'Received expected exception: {e}')
+        #traceback.print_exc(file=sys.stdout)
+        if rebased:
+            assert str(e) == 'page is None'
+        else:
+            assert str(e) == 'orphaned object: parent is None'
+    else:
+        assert 0, 'Did not receive expected exception.'
+    page = None
+    fd4 = next_fd()
+    print(f'{fd1=} {fd2=} {fd3=} {fd4=}')
+    print(f'{document=}')
+    assert fd2 == fd1 + 1
+    assert fd3 == fd1
+    assert fd4 == fd1
+
+def test_xml():
+    path = os.path.abspath(f'{__file__}/../../tests/resources/2.pdf')
+    with fitz.open(path) as document:
+        document.get_xml_metadata()
+
+def test_3112_set_xml_metadata():
+    document = fitz.Document()
+    document.set_xml_metadata('hello world')
+
+def test_archive_3126():
+    if not hasattr(fitz, 'mupdf'):
+        print(f'Not running because known to fail with classic.')
+        return
+    p = os.path.abspath(f'{__file__}/../../tests/resources')
+    p = pathlib.Path(p)
+    archive = fitz.Archive(p)
+    
+def test_3140():
+    if not hasattr(fitz, 'mupdf'):
+        print(f'Not running test_3140 on classic, because Page.insert_htmlbox() not available.')
+        return
+    css2 = ''
+    path = os.path.abspath(f'{__file__}/../../tests/resources/2.pdf')
+    oldfile = os.path.abspath(f'{__file__}/../../tests/test_3140_old.pdf')
+    newfile = os.path.abspath(f'{__file__}/../../tests/test_3140_new.pdf')
+    import shutil
+    shutil.copy2(path, oldfile)
+    def next_fd():
+        fd = os.open(path, os.O_RDONLY)
+        os.close(fd)
+        return fd
+    fd1 = next_fd()
+    with fitz.open(oldfile) as doc:  # open document
+        page = doc[0]
+        rect = fitz.Rect(130, 400, 430, 600)
+        CELLS = fitz.make_table(rect, cols=3, rows=5)
+        shape = page.new_shape()  # create Shape
+        for i in range(5):
+            for j in range(3):
+                qtext = "<b>" + "Ques #" + str(i*3+j+1) + ": " + "</b>"
+                atext = "<b>" + "Ans:" + "</b>"
+                qtext = qtext + '<br>' + atext
+                shape.draw_rect(CELLS[i][j])  # draw rectangle
+                page.insert_htmlbox(CELLS[i][j], qtext, css=css2, scale_low=0)
+        shape.finish(width=2.5, color=fitz.pdfcolor["blue"], )
+        shape.commit()  # write all stuff to the page
+        doc.subset_fonts()
+        doc.ez_save(newfile)
+    fd2 = next_fd()
+    assert fd2 == fd1, f'{fd1=} {fd2=}'
+    os.remove(oldfile)
