@@ -8,6 +8,9 @@ import fitz
 
 def test_rect():
     assert tuple(fitz.Rect()) == (0, 0, 0, 0)
+    if hasattr(fitz, 'mupdf'):
+        assert tuple(fitz.Rect(y0=12)) == (0, 12, 0, 0)
+        assert tuple(fitz.Rect(10, 20, 100, 200, x1=12)) == (10, 20, 12, 200)
     p1 = fitz.Point(10, 20)
     p2 = fitz.Point(100, 200)
     p3 = fitz.Point(150, 250)
@@ -129,6 +132,10 @@ def test_inversion():
 
 def test_matrix():
     assert tuple(fitz.Matrix()) == (0, 0, 0, 0, 0, 0)
+    assert tuple(fitz.Matrix(90)) == (0, 1, -1, 0, 0, 0)
+    if hasattr(fitz, 'mupdf'):
+        assert tuple(fitz.Matrix(c=1)) == (0, 0, 1, 0, 0, 0)
+        assert tuple(fitz.Matrix(90, e=5)) == (0, 1, -1, 0, 5, 0)
     m45p = fitz.Matrix(45)
     m45m = fitz.Matrix(-45)
     m90 = fitz.Matrix(90)
@@ -279,12 +286,7 @@ def test_algebra():
     except:
         failed = True
     assert failed
-    failed = False
-    try:
-        m in r
-    except:
-        failed = True
-    assert failed
+    assert m not in r
 
 
 def test_quad():
@@ -333,3 +335,7 @@ def test_pageboxes():
     for key in keys:
         assert doc.xref_get_key(page.xref, key) == ("array", "[100 142 400 642]")
     assert page.cropbox == page.artbox == page.bleedbox == page.trimbox
+
+def test_3163():
+    b = {'number': 0, 'type': 0, 'bbox': (403.3577880859375, 330.8871765136719, 541.2731323242188, 349.5766296386719), 'lines': [{'spans': [{'size': 14.0, 'flags': 4, 'font': 'SFHello-Medium', 'color': 1907995, 'ascender': 1.07373046875, 'descender': -0.26123046875, 'text': 'Inclusion and diversity', 'origin': (403.3577880859375, 345.9194030761719), 'bbox': (403.3577880859375, 330.8871765136719, 541.2731323242188, 349.5766296386719)}], 'wmode': 0, 'dir': (1.0, 0.0), 'bbox': (403.3577880859375, 330.8871765136719, 541.2731323242188, 349.5766296386719)}]}
+    bbox = fitz.IRect(b["bbox"])
